import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import pairwise_distances
import networkx as nx
import qnorm
import dgl
import torch
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from scipy.sparse import coo_matrix
import random
import numpy as np
import torch_geometric.transforms as GT
from torch_geometric.data import Data
from collections import defaultdict
from heapq import heapify, heappush, heappop


random_seed = 42
random.seed(random_seed)  
np.random.seed(random_seed )
torch.manual_seed(random_seed )
torch.cuda.manual_seed(random_seed)
torch.cuda.manual_seed_all(random_seed )
torch.backends.cudnn.benchmark = False
torch.backends.cudnn.deterministic = True
dgl.random.seed(42)
torch.use_deterministic_algorithms(True) 
torch.backends.cudnn.enabled = False 
import numpy as np
import pandas as pd
import networkx as nx
from sklearn.metrics.pairwise import pairwise_distances
from .CellEmbeddings1 import Node2Vec
import networkx as nx
import qnorm


def _calculate_distance(X, num_jobs=5):
    D = pairwise_distances(X, n_jobs=num_jobs, metric='euclidean')

    return D


def GraphBuild(input_file, k=3):
    nor_file=input_file
    distance_matrix = _calculate_distance(nor_file)
    adj_matrix = np.zeros((distance_matrix.shape[0], distance_matrix.shape[0]))
    for i in range(distance_matrix.shape[0]):
        adj_matrix[np.argsort(distance_matrix[:, i])[1:k + 1], i] = 1
        adj_matrix[i, np.argsort(distance_matrix[:, i])[1:k + 1]] = 1
    adj_df = pd.DataFrame(adj_matrix)
    adj_df.index = input_file.index.tolist()
    adj_df.columns = input_file.index.tolist()
    return adj_df


def trainCellEmbeddings(Graph, Seed=0, Workers=1):
    G = nx.from_pandas_adjacency(Graph)
    Cell_node2vec = Node2Vec(G, dimensions=128, p=4, q=2, walk_length=5, num_walks=20, workers=Workers, seed=Seed)
    tmodel = Cell_node2vec.fit(window=3, epochs=3)
    emb = Cell_node2vec.get_embeddings(tmodel)
    return emb

def knn_graph(data_z):
    data = data_z.copy()
    pos = np.array(data)
    data1 = Data(pos=torch.tensor(pos, dtype=torch.float)
                 )
    graph_transform = GT.KNNGraph(k=5)
    data4 = graph_transform(data1)
    mm = data4.edge_index
    mn = np.array(mm)
    print(data4.edge_index)
    llo = pd.DataFrame(columns=data.index, index=data.index)
    data1 = llo.copy()
    data1.replace(np.nan, 0, inplace=True)
    for i in range(mn.shape[1]):
        l1 = mn[0, i]
        l2 = mn[1, i]
        data1.iloc[l1, l2] = 1
        data1.iloc[l2, l1] = 1
    return data1

def Prim_l(nodes, edges):
    '''  基于最小堆实现的Prim算法  '''
    element = defaultdict(list)
    for start, stop, weight in edges:
        element[start].append((weight, start, stop))
        element[stop].append((weight, stop, start))
    all_nodes = set(nodes)
    used_nodes = set(nodes[0])
    usable_edges = element[nodes[0]][:]
    heapify(usable_edges)
    MST = []
    while usable_edges and (all_nodes - used_nodes):
        weight, start, stop = heappop(usable_edges)
        if stop not in used_nodes:
            used_nodes.add(stop)
            MST.append((start, stop, weight))
            for member in element[stop]:
                if member[2] not in used_nodes:
                    heappush(usable_edges, member)

    return MST

def prim_graph(data_z):
    X=data_z.copy()
    num_jobs =5
    distance_matrix = pairwise_distances(X, n_jobs=num_jobs, metric='euclidean') 
    distance_matrix = pd.DataFrame(distance_matrix)
    m = distance_matrix.index.tolist()
    n = distance_matrix.columns.tolist()
    m_new = [str(i) for i in m]
    n_new = [str(i) for i in n]
    eg = []
    for i in range(len(n)):

        for j in range(len(m)):
            mm = []
            ll00 = m_new[j]
            ll01 = n_new[i]
            ll02 = distance_matrix.iloc[j, i]
            mm.append(ll00)
            mm.append(ll01)
            mm.append(ll02)
            eg.append(mm)
    ll = Prim_l(m_new, eg)
    import numpy as np
    llo = pd.DataFrame(columns=n, index=m)
    data1 = llo.copy()
    data1.replace(np.nan, 0, inplace=True)
    data2 = data1.copy()
    for i in range(len(ll)):
        m = ll[i]
        m0 = int(m[0])
        m1 = int(m[1])

        data2.loc[m0, m1] = 1
        data2.loc[m1, m0] = 1
    return data2


def emd_graph(data):
    AdjGraph=GraphBuild(data)
    embeddingfile=trainCellEmbeddings(AdjGraph)
    prediction_distance = pairwise_distances(embeddingfile.T, n_jobs=5, metric="euclidean")
    adj_matrix_merge = np.array(AdjGraph)

    for j in range(adj_matrix_merge.shape[0]):
        adj_matrix_merge[np.argsort(prediction_distance[:, j])[0:np.sum(adj_matrix_merge[:, j] == 1)], j] = 1

    adj_matrix_merge = pd.DataFrame(adj_matrix_merge)
    adj_matrix_merge.index = data.index.tolist()
    adj_matrix_merge.columns = data.index.tolist()


    return adj_matrix_merge
def Gbuild(data):


    data0=knn_graph(data)
    data1=prim_graph(data)
    data2=emd_graph(data)
    data4=data0+data1+data2

    data_z=data4.copy()
    data_z[data_z<2]=0
    data_z[data_z >= 2] = 1




    qq = coo_matrix(data_z)
    g = dgl.from_scipy(qq)

    return g




def preprocess(graph):
    feat = graph.ndata["feat"]
    graph = dgl.to_bidirected(graph)
    graph.ndata["feat"] = feat

    graph = graph.remove_self_loop().add_self_loop()
    graph.create_formats_()

    return graph




def data_processing(rawfile):
    data=rawfile.copy()

    data1 = np.array(data)
    data1 = pd.DataFrame(data1)


    graph= Gbuild(data1)

    feat = torch.from_numpy(np.array(data1))

    graph.ndata["feat"] = feat

    num_features = len(data.columns.tolist())


    return graph,num_features
