from .edcoder import PreModel


def build_model(args):
    num_heads = args.num_heads
    num_out_heads = args.num_out_heads
    num_hidden = args.num_hidden
    num_layers = args.num_layers
    residual = args.residual
    attn_drop = args.attn_drop
    in_drop = args.in_drop
    norm = args.norm
    negative_slope = args.negative_slope
    encoder_type = args.encoder
    decoder_type = args.decoder
    drop_edge_rate = args.drop_edge_rate
    activation = args.activation
    alpha_l = args.alpha_l
    concat_hidden = args.concat_hidden
    num_features = args.num_features


    model = PreModel(
        in_dim=num_features,
        num_hidden=num_hidden,
        num_layers=num_layers,
        nhead=num_heads,
        nhead_out=num_out_heads,
        activation=activation,
        feat_drop=in_drop,
        attn_drop=attn_drop,
        negative_slope=negative_slope,
        residual=residual,
        encoder_type=encoder_type,
        decoder_type=decoder_type,
        norm=norm,
        drop_edge_rate=drop_edge_rate,
        alpha_l=alpha_l,
        concat_hidden=concat_hidden,
    )


    return model

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter

class IDEC_process(nn.Module):

    def __init__(self,args):
        super(IDEC_process, self).__init__()
        self.alpha = 1.0
        num_heads = args.num_heads
        num_out_heads = args.num_out_heads
        num_hidden = args.num_hidden
        num_layers = args.num_layers
        residual = args.residual
        attn_drop = args.attn_drop
        in_drop = args.in_drop
        norm = args.norm
        negative_slope = args.negative_slope
        encoder_type = args.encoder
        decoder_type = args.decoder
        drop_edge_rate = args.drop_edge_rate
        activation = args.activation
        alpha_l = args.alpha_l
        concat_hidden = args.concat_hidden
        num_features = args.num_features
        n_clusters=args.n_cluster
        self.ae = PreModel(
            in_dim=num_features,
            num_hidden=num_hidden,
            num_layers=num_layers,
            nhead=num_heads,
            nhead_out=num_out_heads,
            activation=activation,
            feat_drop=in_drop,
            attn_drop=attn_drop,
            negative_slope=negative_slope,
            residual=residual,
            encoder_type=encoder_type,
            decoder_type=decoder_type,
            norm=norm,
            drop_edge_rate=drop_edge_rate,
            alpha_l=alpha_l,
            concat_hidden=concat_hidden)
        self.cluster_layer = Parameter(torch.Tensor(n_clusters, num_hidden))
        torch.nn.init.xavier_normal_(self.cluster_layer.data)

    def forward(self, graph, x):

        x_bar, z = self.ae(graph, x)
        q = 1.0 / (1.0 + torch.sum(torch.pow(z.unsqueeze(1) - self.cluster_layer, 2), 2) /  self.alpha)
        q = q.pow((self.alpha + 1.0) / 2.0)
        q = (q.t() / torch.sum(q, 1)).t()
        return x_bar, q



